"""
Merkle Tree Model
"""

from typing import Callable, List
from functools import reduce

from merkly.node import Node, Side
from merkly.utils import (
    validate_hash_function,
    is_power_2,
    slice_in_pairs,
    keccak,
    half,
    validate_leafs,
)


class MerkleTree:
    """
    # 🌳 Merkle Tree implementation

    ## Args:
        - leaves: List of raw data
        - hash_function (Callable[[bytes, bytes], bytes], optional): Function that hashes the data.
            * Defaults to `keccak` if not provided
    """

    def __init__(
        self,
        leaves: List[str],
        hash_function: Callable[[bytes, bytes], bytes] = lambda x, y: keccak(x + y),
    ) -> None:
        validate_leafs(leaves)
        validate_hash_function(hash_function)
        self.hash_function: Callable[[bytes, bytes], bytes] = hash_function
        self.raw_leaves: List[str] = leaves
        self.leaves: List[str] = self.__hash_leaves(leaves)
        self.short_leaves: List[str] = self.short(self.leaves)

    def __hash_leaves(self, leaves: List[str]) -> List[str]:
        return list(map(lambda x: self.hash_function(x.encode(), bytes()), leaves))

    def __repr__(self) -> str:
        return f"""MerkleTree(\nraw_leaves: {self.raw_leaves}\nleaves: {self.leaves}\nshort_leaves: {self.short(self.leaves)})"""

    def short(self, data: List[str]) -> List[str]:
        return [x[:2] for x in data]

    @property
    def root(self) -> bytes:
        return self.make_root(self.leaves)

    def proof(self, raw_leaf: str) -> List[Node]:
        return self.make_proof(
            self.leaves, [], self.hash_function(raw_leaf.encode(), bytes())
        )

    def verify(self, proof: List[bytes], raw_leaf: str) -> bool:
        full_proof = [self.hash_function(raw_leaf.encode(), bytes())]
        full_proof.extend(proof)

        def concat_nodes(left: Node, right: Node) -> Node:
            if isinstance(left, Node) is not True:
                start_node = left
                if right.side == Side.RIGHT:
                    data = self.hash_function(start_node, right.data)
                    return Node(data=data, side=Side.LEFT)
                else:
                    data = self.hash_function(right.data, start_node)
                    return Node(data=data, side=Side.RIGHT)
            else:
                if right.side == Side.RIGHT:
                    data = self.hash_function(left.data, right.data)
                    return Node(data=data, side=Side.LEFT)
                else:
                    data = self.hash_function(right.data, left.data)
                    return Node(data=data, side=Side.RIGHT)

        return reduce(concat_nodes, full_proof).data == self.root

    def make_root(self, leaves: List[bytes]) -> bytes:
        if len(leaves) == 0:
            raise ValueError("Cannot get root of an empty tree")

        while len(leaves) > 1:
            next_level = []
            for i in range(0, len(leaves) - 1, 2):
                next_level.append(self.hash_function(leaves[i], leaves[i + 1]))

            if len(leaves) % 2 == 1:
                next_level.append(leaves[-1])

            leaves = next_level

        return leaves[0]

    def make_proof(
        self, leaves: List[bytes], proof: List[Node], leaf: bytes
    ) -> List[Node]:
        """
        # Make a proof

        ## Dev:
            - if the `leaf` index is less than half the size of the `leaves`
        list then the right side must reach root and vice versa

        ## Args:
            - leaves: List of leaves
            - proof: Accumulated proof
            - leaf: Leaf for which to create the proof

        ## Returns:
            - List of Nodes representing the proof
        """

        try:
            index = leaves.index(leaf)
        except ValueError as err:
            msg = f"Leaf: {leaf} does not exist in the tree: {leaves}"
            raise ValueError(msg) from err

        if is_power_2(len(leaves)) is False:
            return self.mix_tree(leaves, [], index)

        if len(leaves) == 2:
            if index == 1:
                proof.append(Node(data=leaves[0], side=Side.LEFT))
            else:
                proof.append(Node(data=leaves[1], side=Side.RIGHT))
            proof.reverse()
            return proof

        left, right = half(leaves)

        if index < len(leaves) / 2:
            proof.append(Node(data=self.make_root(right), side=Side.RIGHT))
            return self.make_proof(left, proof, leaf)
        else:
            proof.append(Node(data=self.make_root(left), side=Side.LEFT))
            return self.make_proof(right, proof, leaf)

    def mix_tree(
        self, leaves: List[bytes], proof: List[Node], leaf_index: int
    ) -> List[Node]:
        if len(leaves) == 1:
            return proof

        if leaf_index % 2 == 0:
            if leaf_index + 1 < len(leaves):
                node = Node(data=leaves[leaf_index + 1], side=Side.RIGHT)
                proof.append(node)
        else:
            node = Node(data=leaves[leaf_index - 1], side=Side.LEFT)
            proof.append(node)

        return self.mix_tree(self.up_layer(leaves), proof, leaf_index // 2)

    def up_layer(self, leaves: List[bytes]) -> List[bytes]:
        new_layer = []
        for pair in slice_in_pairs(leaves):
            if len(pair) == 1:
                new_layer.append(pair[0])
            else:
                data = self.hash_function(pair[0], pair[1])
                new_layer.append(data)
        return new_layer



"""
Utils functions
"""

from typing import Callable, List, Tuple
from Crypto.Hash import keccak as cryptodome_keccak
import types


class PowerOfTwoError(Exception):
    def __init__(self, number):
        self.number = number
        super().__init__(
            f"Size of leafs should be a power of 2 your leafs length is: {number}"
        )


class InvalidHashFunctionError(Exception):
    """Exception raised for invalid hash function."""

    def __init__(self) -> None:
        self.message = "Must type of: (bytes, bytes) -> bytes"
        super().__init__(self.message)


def keccak(data: bytes) -> bytes:
    """
    # Hash `data: str` using keccak256
    - params `data: str`
    - return `hexadecimal: str`

    ```python
    >>> keccak("merkle")
    "326fe0d8a70ab934a7bf9d1323c6d87ee37bbe70079f82e72203b1e07c0c185c"

    >>> keccak("tree")
    "b2510336c6497719adadc7ade198c988520f3349445f074dc729df0f3c2b12ad"

    >>> keccak("bitcoin")
    "7dee6e1aa550de37364ec77e03e62ea56bf42037b8297280de9d844d88444e4d"

    >>> keccak("ethereum")
    "541111248b45b7a8dc3f5579f630e74cb01456ea6ac067d3f4d793245a255155"
    ```
    """

    keccak_256 = cryptodome_keccak.new(digest_bits=256)
    keccak_256.update(data)

    return keccak_256.digest()


def half(list_item: List[int]) -> Tuple[int, int]:
    """
    # Slice a `x: list[int]` in a pairs
    - params `x: list[int]`
    - return `list: list[list[int]]

    ```python
    >>> slicer([1,2,3,4])
    [[1, 2], [3, 4]]

    >>> slicer([i for i in range(10)])
    [[0, 1, 2, 3, 4], [5, 6, 7, 8, 9]]
    ```
    """

    length = len(list_item) // 2

    return (list_item[:length], list_item[length:])


def slice_in_pairs(list_item: list):
    """
    # Slice a `x: list[int]` in pairs, pairs is sublist of 2 items
    - params `x: list[int]`
    - return `list: list[list[int]]

    ```python
    >>> slicer([1,2,3,4])
    [[1, 2], [3, 4]]

    >>> slicer([i for i in range(10)])
    [[0, 1], [2, 3], [4, 5], [6, 7], [8, 9]]
    ```
    """

    return [list_item[i : i + 2] for i in range(0, len(list_item), 2)]


def validate_leafs(leafs: List[str]):
    size = len(leafs)

    if size < 2:
        raise Exception("Invalid size, need > 2")

    a = isinstance(leafs, List)
    b = all(isinstance(leaf, str) for leaf in leafs)
    if not (a and b):
        raise Exception("Invalid type of leafs")


def validate_hash_function(hash_function: Callable[[bytes, bytes], bytes]):
    a = isinstance(hash_function, types.FunctionType)
    b = callable(hash_function)
    try:
        c = isinstance(hash_function(bytes(), bytes()), bytes)
    except TypeError:
        c = False

    valid = a and b and c
    if not valid:
        raise InvalidHashFunctionError()


def is_power_2(number: int) -> bool:
    """
    # Verify if `x: int` is power of 2
    - params `x: int`
    - return `bool`

    ```python
    assert is_power_2(2) == True
    assert is_power_2(3) == False
    assert is_power_2(16) == True
    assert is_power_2(900) == False
    ```
    """

    left: bool = number & (number - 1) == 0
    right: bool = number != 0
    if left and right:
        return True
    else:
        return False
